"""
Author: Wenyu Ouyang
Date: 2022-12-02 10:42:19
LastEditTime: 2023-10-27 14:53:16
LastEditors: Wenyu Ouyang
Description: Top-level package for hydroutils.
FilePath: /hydroutils/hydroutils/__init__.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

__author__ = """Wenyu Ouyang"""
__email__ = 'wenyuouyang@outlook.com'
__version__ = '0.0.10'

from .hydro_log import *
